/*	$NetBSD$	*/

/*
 * Copyright (c) 2001 ITOH Yasufumi.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/types.h>
#include <sys/endian.h>
#include <string.h>

#include "dcmem.h"

/*
 * find a directory entry
 *
 * error: (none)
 */
struct dcdirent *
dcmem_dir_enumerate(m, ppos)
	const struct dcmem *m;
	int *ppos;		/* in/out: position in directory (0: top) */
{
	int pos = *ppos;
	struct dcdirent *d, *dend;

	dend = m->dm_dir + m->dm_dirsz * (DCMEM_BLK_SZ / sizeof *d);

	for (d = m->dm_dir + pos; d < dend; d++) {
		pos++;
		if (d->dd_type != DCMEM_FTYPE_NOFILE) {
			*ppos = pos;
			return d;
		}
	}
	*ppos = pos;
	return NULL;
}

/*
 * look-up a file by name
 *
 * error: (none)
 */
struct dcdirent *
dcmem_dir_lookup_name(m, name)
	const struct dcmem *m;
	const u_int8_t *name;
{
	struct dcdirent *d;
	int pos;

	/* start from the top position */
	pos = 0;

	while ((d = dcmem_dir_enumerate(m, &pos)) != NULL) {
		if (!strncmp((const char *) d->dd_name, (const char *) name,
		    sizeof d->dd_name))
			break;
	}

	return d;
}

int
dcmem_dir_remove_name(m, name)
	struct dcmem *m;
	const u_int8_t *name;
{
	struct dcdirent *d;
	int cnt;
	int pos;

	cnt = 0;
	pos = 0;	/* start from the top position */

	while ((d = dcmem_dir_enumerate(m, &pos)) != NULL) {
		if (!strncmp((const char *) d->dd_name, (const char *) name,
		    sizeof d->dd_name)) {
			dcmem_fat_free_chain_data(m, le16toh(d->dd_start));
			d->dd_type = DCMEM_FTYPE_NOFILE;
			cnt++;
		}
	}

	return cnt;
}

/*
 * find a free directory slot and write the dirent into the slot
 *
 * errors:	DCMEM_EEXISTS
 *		DCMEM_EDIRFULL
 */
struct dcdirent *
dcmem_dir_alloc(m, name, perr)
	struct dcmem *m;
	const u_int8_t *name;
	int *perr;
{
	struct dcdirent *d, *dend;

	if ((d = dcmem_dir_lookup_name(m, name)) != 0) {
		/* file already exists */
		*perr = DCMEM_EEXISTS;
		return NULL;
	}

	/*
	 * look for free slot
	 */
	dend = m->dm_dir + m->dm_dirsz * (DCMEM_BLK_SZ / sizeof *d);
	for (d = m->dm_dir; d < dend; d++) {
		if (d->dd_type == DCMEM_FTYPE_NOFILE) {
			return d;
		}
	}

	/* no free slot found */
	*perr = DCMEM_EDIRFULL;
	return NULL;
}
