/*	$NetBSD$	*/

/*
 * Copyright (c) 2001 ITOH Yasufumi.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/types.h>
#include <sys/endian.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#ifdef DCMEM_DEBUG
#include <err.h>
#endif

#include "dcmem.h"

static int dcmem_read_rblk __P((struct dcmem *));
static int dcmem_read_fat __P((struct dcmem *));
static int dcmem_read_dir __P((struct dcmem *));

int
dcmem_read_1(m, ptr, blk)
	const struct dcmem *m;
	void *ptr;
	int blk;
{
	u_int8_t buf[DCMEM_BLK_SZ];
	int retrycnt = DCMEM_RETRYCNT;
	off_t off = blk * DCMEM_BLK_SZ;
	int fd = m->dm_fd;
	int r, s = DCMEM_EHWPROBLEM;
	int first;

	/* make sure we read at least twice */
	first = 1;

	for (;;) {
	retry0:
		if (retrycnt-- == 0)
			break;

		/* read and compare */
		memset(ptr, 0, DCMEM_BLK_SZ);
		if ((r = pread(fd, ptr, DCMEM_BLK_SZ, off)) < 0) {
			s = DCMEM_ESYS;
			goto retry0;
		} else if (r != DCMEM_BLK_SZ) {
			s = DCMEM_EUXEOF;
			goto retry0;
		}
		s = DCMEM_EHWPROBLEM;
		if (!first && memcmp(ptr, buf, DCMEM_BLK_SZ) == 0)
			return 0;
		first = 0;

	retry1:
		if (retrycnt-- == 0)
			break;

		/* read and compare */
		memset(buf, 0xff, DCMEM_BLK_SZ);
		if ((r = pread(fd, buf, DCMEM_BLK_SZ, off)) < 0) {
			s = DCMEM_ESYS;
			goto retry1;
		} else if (r != DCMEM_BLK_SZ) {
			s = DCMEM_EUXEOF;
			goto retry1;
		}
		s = DCMEM_EHWPROBLEM;
		if (memcmp(ptr, buf, DCMEM_BLK_SZ) == 0)
			return 0;
	}

	return s;
}

static int
dcmem_read_rblk(m)
	struct dcmem *m;
{
	int st;

	/* should use GETMINFO to get root block location? */

	m->dm_rblk_changed = 0;
	if ((st = dcmem_read_1(m, &m->dm_rblk, DCMEM_DEFAULT_ROOTBLK_LOC))
	    != 0) {
#ifdef DCMEM_DEBUG
		warn("dcmem_read_rblk: read failed");
#endif
		return st;
	 }

	if (memcmp(m->dm_rblk.dr_mark, DCMEM_RB_MARK,
	    sizeof m->dm_rblk.dr_mark))
		return DCMEM_EFORMAT;

	m->dm_maxblk = le16toh(m->dm_rblk.dr_maxblk);
	m->dm_minblk = le16toh(m->dm_rblk.dr_minblk);
	m->dm_rbloc = le16toh(m->dm_rblk.dr_rbloc);
	m->dm_fatloc = le16toh(m->dm_rblk.dr_fatloc);
	m->dm_fatsz = le16toh(m->dm_rblk.dr_fatsz);
	m->dm_dirloc = le16toh(m->dm_rblk.dr_dirloc);
	m->dm_dirsz = le16toh(m->dm_rblk.dr_dirsz);
	m->dm_nblk = le16toh(m->dm_rblk.dr_nblk);

	/*
	 * simple consistency check
	 */
	if (m->dm_maxblk <= m->dm_minblk)
		return DCMEM_ERBLK;
	if (m->dm_rbloc != DCMEM_DEFAULT_ROOTBLK_LOC)
		return DCMEM_ENOTSUPP;
	if (m->dm_fatloc + m->dm_fatsz > m->dm_maxblk || m->dm_fatsz == 0)
		return DCMEM_ERBLK;
	if (m->dm_dirloc > m->dm_maxblk || m->dm_dirsz == 0)
		return DCMEM_ERBLK;
	if (m->dm_nblk == 0)
		return DCMEM_ERBLK;

	/* XXX for now */
	if (m->dm_fatsz != 1)
		return DCMEM_ENOTSUPP;

	/*
	 * allocate and read FAT
	 */
	if ((m->dm_fat = malloc((size_t) (m->dm_fatsz * DCMEM_BLK_SZ))) == NULL)
		return DCMEM_ESYS;

	if ((st = dcmem_read_fat(m)) != 0) {
		free(m->dm_fat);
		return st;
	}

	/*
	 * allocate and read directory
	 */
	if ((m->dm_dir = malloc((size_t) (m->dm_dirsz * DCMEM_BLK_SZ)))
	    == NULL) {
		free(m->dm_fat);
		return DCMEM_ESYS;
	}

	if ((st = dcmem_read_dir(m)) != 0) {
		free(m->dm_fat);
		free(m->dm_dir);
		return st;
	}

	return 0;
}

struct dcmem *
dcmem_open(fn, flags, perr)
	const char *fn;
	int flags;
	int *perr;
{
	struct dcmem *m;
	int s;

	if ((m = malloc(sizeof(struct dcmem))) == NULL) {
		*perr = DCMEM_ESYS;
		return (struct dcmem *) 0;
	}

	m->dm_flags = flags;
	m->dm_writing = 0;

	if ((m->dm_fd =
	    open(fn, flags & DCMEM_FRDONLY ? O_RDONLY : O_RDWR)) < 0) {
		s = DCMEM_ESYS;
		goto out;
	}

	if ((s = dcmem_read_rblk(m)) != 0)
		goto out;

	return m;

out:
	*perr = s;
	free(m);
	return (struct dcmem *) 0;
}

static int
dcmem_read_fat(m)
	struct dcmem *m;
{

	/* XXX for now */
	if (m->dm_fatsz != 1)
		return DCMEM_ENOTSUPP;

	return dcmem_read_1(m, m->dm_fat, m->dm_fatloc);
}

int
dcmem_read_chain(m, buf, top, sz, is_sys)
	const struct dcmem *m;
	void *buf;
	int top;	/* fat # of top of chain */
	int sz;		/* size in blocks (for consistency check) */
	int is_sys;
{
	int i;
	int st;
	char *p = buf;

	for (i = top; sz > 0; sz--, p += DCMEM_BLK_SZ) {
		/* check for FAT consistency */
		if (is_sys) {
			if (i < m->dm_nblk || i >= m->dm_maxblk)
				return DCMEM_EINCONSISTENT;
		} else {
			if (i < 0 || i >= m->dm_nblk)
				return DCMEM_EINCONSISTENT;
		}

		/* read it */
		if ((st = dcmem_read_1(m, p, i)) != 0) {
#ifdef DCMEM_DEBUG
			warn("dcmem_read_chain: read failed: blk %d", i);
#endif
			return st;
		}

		i = dcmem_fat_next(m, i);
	}

	/* FAT reaches at the end? */
	if (i != -1)
		return DCMEM_EINCONSISTENT;

	return 0;
}

static int
dcmem_read_dir(m)
	struct dcmem *m;
{

	return dcmem_read_chain(m, m->dm_dir, m->dm_dirloc, m->dm_dirsz, 1);
}

int
dcmem_close(m)
	struct dcmem *m;
{
	int st;

	st = dcmem_flush(m);

	free(m->dm_dir);
	free(m->dm_fat);
	free(m);

	return st;
}
